// Copyright (c) 2017, Perforce Software, Inc.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1.  Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//
// 2.  Redistributions in binary form must reproduce the above copyright
//     notice, this list of conditions and the following disclaimer in the
//     documentation and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE SOFTWARE, INC. BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

RCSCHANGE = "$Change: 1594901 $";
RCSDATE = "$Date: 2017/11/29 $";

p4 = "C:\\Program Files\\Perforce\\p4.exe";
p4d = "C:\\Program Files\\Perforce\\Server\\p4d.exe";

var oFSO = new ActiveXObject("Scripting.FileSystemObject");
p4dname = oFSO.GetFileName(p4d);

scp = "C:\\cygwin\\bin\\scp.exe";
ssh = "C:\\cygwin\\bin\\ssh.exe";
rm = "rm";
chmod = "chmod";
perms = "500";

f = WScript.ScriptName;
BENCHMARK = f.split(".") [0];
CHILD = BENCHMARK + "child" ;

BINPATH="bin." ;

DATETIME = getdatestr();

//
// Host entries are of the form:
//
//   <host>=<platform>[,[<children>][,[<browses>][,[<connection>][,[<seed>][,[<user-offset>]]]]]]
//
HOSTS="client1name=platform,32,1024 client2name=platform,32,1024";

LICENSE= "C:\\p4licenses\\license" ;

P4PORT = "servername:31697"; // must NOT be an existing server!

// must NOT be an existing installation!
P4ROOT = "C:\\p4dbs\\" + BENCHMARK ;

P4JOURNALdir = "C:\\p4journals\\" + BENCHMARK;
// must NOT be an existing journal!
P4JOURNAL = P4JOURNALdir  + "\\journal" ;

P4LOGdir = "C:\\p4logs\\" + BENCHMARK ;
// must NOT be an existing log!
P4LOG = P4LOGdir + "\\log" ;

SIDENTITY="browse" ;
SOPTIONS="";
SOPTIONS=SOPTIONS  + " -o BatchMode=yes";
SOPTIONS=SOPTIONS  + " -o LogLevel=ERROR";
SOPTIONS=SOPTIONS  + " -o StrictHostKeyChecking=no";
SOPTIONS=SOPTIONS  + " -o UserKnownHostsFile=NUL\\:";

CYGWINVARS = "nodosfilewarning";

TMPPOSIX="/tmp/" ;
TMPWINDOWS = "\$TMP/" ;

YAMLFILE = BENCHMARK + "." + DATETIME + ".yml";
YAMLINDENT = "    ";

ZCKP = "C:\\p4ckps\\reference01.2017.2.ckp.gz";

// Extract certain values from the Windows environment.
var oShell = new ActiveXObject("WScript.Shell");
var uname = oShell.ExpandEnvironmentStrings("%COMPUTERNAME%").split(".") [0];
var SUSER = oShell.ExpandEnvironmentStrings("%USERNAME%");
var os = oShell.ExpandEnvironmentStrings("%OS%");
var arch = oShell.ExpandEnvironmentStrings("%PROCESSOR_ARCHITECTURE%");
var localtmp = oShell.ExpandEnvironmentStrings("%TMP%");
var p4portstr = P4PORT.split(":");
var p4portnum = parseInt(p4portstr[1]);

prtdebug = 0 ;

function logmessage(msg)
{

	WScript.Stdout.Write(msg);
}

function debugmessage(msg)
{

	if (prtdebug)
	{
	    WScript.Stdout.Write("\nDEBUG: " + msg);
	}
}

function getrcsvalue(rcsstr)
{
	//
	// Return an RCS value from a string. The RCS value is the
	// substring between the first and last space characters.
	//
	var match = new RegExp(/^[^ ]* (.*) [^ ]*$/);
        return rcsstr.replace(match,"$1");
}

function createyamlfile()
{
	var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	var outfile = oFSO.GetFolder(".") + "\\" + YAMLFILE;
	oFSO.CreateTextFile(outfile);
	return outfile;
}

function writeyamlfile(yamlfile, yamllevel, str)
{
	ForAppending = 8; 

	var i = yamllevel;

	var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	var outfile = oFSO.OpenTextFile(yamlfile, ForAppending, false);

	while ( i != 0 )
	{
	    outfile.Write(YAMLINDENT);
	    i--;
	}

	outfile.WriteLine(str);
	outfile.Close();
}

function chgcase(str)
{
	//
	// Change the first byte of each word in a string to uppercase and
	// change the remaining bytes of each word to lowercase. Whitespace
	// between words is preserved. Examples of the behavior of this
	// function are:
	//
	//   chgcase("foobar") returns: Foobar
	//   chgcase("foo bAR") returns: Foo Bar
	//   chgcase("foo  bA/R") returns: Foo  Ba/r
	//
	return str.replace(/\w\S*/g, function(txt){return txt.charAt(0).toUpperCase() + 
	    txt.substr(1).toLowerCase();});
}

function getversion(prog)
{
	var p4cmd =  prog + " -V";

	var re = new RegExp(/^Rev. (.*)$/);

	oShell = new ActiveXObject("WScript.Shell");
	oExec = oShell.Exec(p4cmd);

	//
	// Read the version string output and peal out everything after Rev
	//
        while (!oExec.Stdout.AtEndOfStream)
	{
	    version = oExec.Stdout.ReadLine();
	    if (re.test(version))
	    {
		return version.replace(re,"$1");

	    } 
	}	
}

function getargs()
{
	var args = "";

	for (var n = 0; n < phase.length; n++)
	{
	    args = args + " " + phase[n];
	}

	return args;
}

function setenv()
{

	// P4USER is the current USERNAME
	var oShell = new ActiveXObject("WScript.Shell");
	var P4USER = oShell.ExpandEnvironmentStrings("%USERNAME%");

	// Create Environment variables for the following.
	oShell.Environment("PROCESS").Item("P4PORT") = P4PORT;
	oShell.Environment("PROCESS").Item("P4USER") = P4USER;
	oShell.Environment("PROCESS").Item("CYGWIN") = CYGWINVARS;
}

function getpid()
{

	// The pid from the cscript process is extracted and used as
	// a unique identifier. If that doesn't exist or is unusable
	// a random number between 0 and 1000 will be used.

	var oWMIService = GetObject("winmgmts:\\\\.\\root\\CIMV2");
	var colProcess = oWMIService.ExecQuery("SELECT Handle, Caption " +
		"FROM Win32_Process WHERE Caption = 'cscript.exe'");

	var numret = 0;
	var progpid = 0;

        var ei = new Enumerator(colProcess);
        for (; !ei.atEnd(); ei.moveNext())
        {
	    var oItem = ei.item();
	    // Increment the number of pids found.
	    numret++;
	
	}

	// If the number of processes found is not equal to 1 we'll 
	// use a random number instead.
	if (numret != 1)
	{
	    progpid = Math.floor(Math.random()*1000);
	} else {
	    progpid = oItem.handle;
	}
	
	return progpid;
}

function getusablemem()
{

	//
	// The Windows sort command works best if available memory
	// is specified on the command line. The amount of total
	// physical memory is determined and 1 GB of memory
	// is subtracted from the total. This is the amount specified
	// on the sort command line.
	//
	var oWMIService = GetObject("winmgmts:\\\\.\\root\\CIMV2");
	var colWMIService = oWMIService.ExecQuery("SELECT " + 
		"TotalPhysicalMemory FROM Win32_ComputerSystem");

	var ei = new Enumerator(colWMIService);
	for (; !ei.atEnd(); ei.moveNext())
	{
	    var oItem = ei.item();
	    var bytememory =  oItem.TotalPhysicalMemory;
	}

	//
	// Change total available physical memory from bytes to KB.
	//
	var kbytememory = (parseInt(bytememory) / 1024);

	//
	// Subtract 1GB from physical memory. We don't want to bring
	// the machine to its knees by taking it all.
	// 
	usablememory = (kbytememory - 1048576);

	return usablememory;
}

function makedir (dirpath)
{
	var oShell = new ActiveXObject("WScript.Shell")
        var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	var parentpath = oFSO.GetParentFolderName(dirpath);

        // Create the parent folder if it doesn't exist
        if (!oFSO.FolderExists(parentpath))
	{
	    makedir(parentpath);
	}
        if (!oFSO.FolderExists(dirpath))
	{
	    oFSO.CreateFolder(dirpath);
       	} 
}

function checkprocess(process)
{

	//
	// Select all the processes from the WMI services and attempt
	// to match the passed in process.
	//

	var ckproc = 1;
	var re = new RegExp(process);

	var oWMIService = GetObject("winmgmts:\\\\.\\root\\CIMV2");
	var colProcess = oWMIService.ExecQuery("SELECT * FROM Win32_Process");

	var ei = new Enumerator(colProcess);
	for (; !ei.atEnd(); ei.moveNext())
	{
	    var oItem = ei.item();
	    if (re.test(oItem.name))
	    {
		ckproc = 0;
	    }
	}

	return ckproc;
}

function threadcount(process)
{

	var tcnt = 0;
	var re = new RegExp(process);
	var oWMIService = GetObject("winmgmts:\\\\.\\root\\CIMV2");
	var colProcess = oWMIService.ExecQuery("select Name, ThreadCount from Win32_Process where name=\"" + process + "\"");
	var ei = new Enumerator(colProcess);
	for (; !ei.atEnd(); ei.moveNext())
	{
	    var oItem = ei.item();
	    debugmessage("Name " + oItem.Name + 
		" Threadcnt " + oItem.threadcount + "\n");
	    tcnt = oItem.threadcount; 
	}
	return tcnt;
}

function getseconds()
{

	var oDate = new Date();
	return oDate.getTime() / 1000; // return the seconds
}

function getelapsed(start)
{
	var end = getseconds();
	return (end - start);
}

function getlocalhost()
{
	return uname.toLowerCase();
}

function getremotehost(raw)
{
        //
        // Parse out and return the host name.
        //
        var hostname =  raw.split("=");
        return hostname[0];
}

function getplatform(raw)
{
        //
        // Parse out and return the platform.
        //

        var match = new RegExp(/^.*=([^,]*).*$/);
        return raw.replace(match,"$1");
}

function getchildren(raw)
{

	//
	// Parse out and return the (default, if unspecified) number of children
	//
	var record = raw.split(",");
	var children = parseInt(record[1]);

	//
	// If the returned children number is No A Number the default is used 
	//
	if (isNaN(children))
	{
	    children = 32;
	}

	return children;	
}

function getoptions(raw)
{
        //
        // Parse out any specified child options. The child will provide
        // the defaults for the unspecified options.
        //
        var opts = raw.split(",");
        browses = parseInt(opts[2]);
        connect = opts[3];
        seed = parseInt(opts[4]);
        return [browses,connect,seed];
}

function getuseroffset(raw)
{
	//
	// Parse out and return the user offset.
	//
	var opts = raw.split(",");
	return parseInt(opts[5]);
}

function overlap(useroffset1, nchildren1, useroffset2, nchildren2)
{
	//
	// Check for non-trivial overlap, where:
	//
	//   useroffset1: beginning of the first range
	//   nchildren1: number in the first range
	//   useroffset2: beginning of the second range
	//   nchildren2: number in the second range
	//
	return useroffset1 < useroffset2 + nchildren2 &&
	    useroffset2 < useroffset1 + nchildren1 &&
	    nchildren1 != 0 && nchildren2 != 0;
}

function usersfit(nhosts, nchildrens, useroffsets)
{
	//
	// Determine if the users specified by the number of children and
	// the user offsets defined for the hosts will result in unique users.
	// Since the users for each pair of hosts need checked only once,
	// only n(n-1)/2 overlap checks are needed.
	//

	var nchildrenihost, useroffsetihost;
	var nchildrenjhost, useroffsetjhost;

	for (var ihost = 1; ihost < nhosts; ihost++)	// start at second host
	{
	    useroffsetihost = useroffsets[ihost];
	    if (isNaN(useroffsetihost))
	    {
		continue;
	    }

	    nchildrenihost = nchildrens[ihost];

	    //
	    // Check the users of this host for overlaps with
	    // the users of only the prior hosts.
	    //
	    var jhost = ihost;
	    while (jhost > 0)
	    {
		jhost--;

		useroffsetjhost = useroffsets[jhost];
		if (isNaN(useroffsetjhost))
		{
		    continue;
		}

		nchildrenjhost = nchildrens[jhost];

		if (overlap(useroffsetihost, nchildrenihost,
		    useroffsetjhost, nchildrenjhost))
		{
		    //
		    // The users defined by the number of children and
		    // the user offsets for these two hosts overlap.
		    //
		    return false;
		}
	    }
	}

	//
	// No overlaps were found.
	//
	return true;
}

function makeuseroffset(nchildrenihost, nhosts, nchildrens, useroffsets)
{
	//
	// Make a user offset that for the passed number of children,
	// there will be no overlap with users of any other host.
	//

	var candidate = 0;	// first candidate user offset

	outer: while (true)	// exit only when user offset doesn't overlap
	{

	    for (var jhost = 0; jhost < nhosts; jhost++)
	    {
		useroffsetjhost = useroffsets[jhost];
		if (isNaN(useroffsetjhost))
		{
		    continue;
		}

		nchildrenjhost = nchildrens[jhost];

		if (overlap(candidate, nchildrenihost,
		    useroffsetjhost, nchildrenjhost))
		{
		    //
		    // The candidate user offset and the passed
		    // number of children overlap the users of this host;
		    // set a new candidate user offset to just after the
		    // users of this host.
		    //
		    candidate = useroffsetjhost + nchildrenjhost;

		    //
		    // Begin the next iteration of the outer loop to
		    // start checking the new candidate user offset.
		    //
		    continue outer;
		}
	    }

	    //
	    // The candidate user offset and the passed number of children
	    // do not overlap the users of any host; exit the outer loop.
	    //
	    break;

	}

	return candidate;	// return the successful candidate user offset
}

function fituseroffsets(nhosts, nchildrens, useroffsets)
{
	//
	// For the hosts in the HOSTS definition that did not have
	// a user offset specified, make user offsets such that
	// all resulting users will be unique.
	//
	for (var ihost = 0; ihost < nhosts; ihost++)
	{
	    if (isNaN(useroffsets[ihost]))
	    {
		useroffsets[ihost] = makeuseroffset(nchildrens[ihost],
		    nhosts, nchildrens, useroffsets);
	    }
	}
}

function checkssh()
{
	var oShell = new ActiveXObject("WScript.Shell");
	var expresult = new RegExp(/^Hello World!/);
	var replyerror = "";
	var bail = 0;

        var hostrecords = HOSTS.split(/ /);

	//
	// Check ssh connectivity between the server and client machines.
	// If there is a lack of connectivity anywhere, bail.
	//
	
	logmessage("Checking ssh connectivity...\n");
	var localhost = getlocalhost();

	for (var r = 0; r < hostrecords.length; r++)
	{
	    var host = getremotehost(hostrecords[r]);

	    logmessage("  between " + localhost + " and " + host + "...");

	    var sshcmd = "\"" + ssh + "\" " + SOPTIONS + " -i" + SIDENTITY + 
		" \"" + SUSER + "@" + host + "\" " + " echo Hello World!" ;

	    debugmessage(sshcmd);

	    var oExec = oShell.exec(sshcmd);

	    if (oExec.ProcessID == 0 && oExec.Status == 1)
	    {
		logmessage("oShell.exec failed.\n");
		bail = 1;
	    }

	    while (!oExec.Stderr.AtEndOfStream)
	    {
		replyerror = oExec.Stderr.ReadLine();
	 	if (replyerror)
		{
		    logmessage(" FAILED!\n");
		    logmessage("    " + replyerror + "\n");
		    bail = 1;
		}
	    }

	    while (!oExec.Stdout.AtEndOfStream)
	    {
		reply = oExec.Stdout.ReadLine();
		if (expresult.test(reply))
		{
		    logmessage( " passed.\n");
		} else {
		    bail = 1;
		}
	    }
	}
	if (bail)
	{
	    logmessage("Correct ssh connectivity before continuing.\n");
	    WScript.Quit(1);
	}
}

function islistening()
{
	//
	// Determine if the server is listening by returning the status
	// from executing the "p4 info" command.
	//
        var oShell = new ActiveXObject("WScript.Shell");
        var oExec = oShell.exec(p4 + " -p " + p4portnum + " info");

        if (oExec.ProcessID == 0 && oExec.Status == 1)
	{
	    logmessage("oShell.exec failed.\n");
	}

	// Wait until "p4 info" is completed until checking ExitCode.
        while (oExec.Status == 0)
	{
	    WScript.Sleep(1000);
	}

	return oExec.ExitCode;
}

function waitlistener()
{
	var cnt = 0;
	var wlisten = 1;
	//
	// Give waitlistener 5 tries before failing with error message.
	//
	do	
	{
	    WScript.Sleep(1000);
	    if (cnt == 5 )
	    {
		logmessage(p4dname + " Failed to Start");
		WScript.Quit(1);
	    }
	    cnt++;
	}
	while (islistening() != 0);

}

function waitstopped(p4dname)
{
	//
	// Wait for the server to completely stop by
	// periodically checking for related p4d threads.
	//
	do
	{
	    var openthreads = threadcount(p4dname);
	    WScript.Sleep(1000);
	    if (cnt == 5)
	    {
		logmessage(p4dname + " Failed to stop");
		WScript.Quit(1);
	    }
	    cnt++;
	}
	while (openthreads != 0);	
	
}

function startserver()
{

	logmessage("Starting Server...");

	p4dargs =  " -d -p " + p4portnum + " -r \"" + P4ROOT + "\"  -J \"" + 
		P4JOURNAL  + "\" -q -L \"" + P4LOG + "\" -v server=1 -v track=0";

	var oShell = new ActiveXObject("WScript.Shell");

	//	
	// Using the run method because p4d needs to be attached to a 
	// spawn window so that it is not affected my control-c during the test.
	//
	var oExec = oShell.run("\"" +  p4d + "\" " + p4dargs,0,false);
	debugmessage("\"" +  p4d + "\" " + p4dargs);

	waitlistener();
		
	//
	// Sleeping for 10 seconds before starting the test 
	//
	WScript.Sleep(10000);
	logmessage(" done.\n");

}

function stopserver()
{
	if (islistening() == 0)
	{
	    //
	    // Stop the server and wait for it to completely stop and quiesce.
	    //	
	    logmessage("Stopping server...");

	    cnt = 0;
	    procstatus = 0;

	    var p4portstr = P4PORT.split(":");
	    var portnum = parseInt(p4portstr[1]);

	    var oShell = new ActiveXObject("WScript.Shell");
	    var oExec = oShell.exec("\"" + p4 + "\" -p " + "127.0.0.1:" + portnum + 
		" admin stop");
	
	    while (oExec.Status == 0)
	    {
		WScript.Sleep(1000);
	    }

	    waitstopped(p4dname);
	    WScript.Sleep(10000);
	    logmessage(" done.\n");
	} else {
	    logmessage("Server not listening. (Already stopped?)\n");
	}	

}

function setup(yamlfile, yamllevel)
{
	//
	// Check ssh connectivity between the server and client machines
	// before any other setup steps are executed.
	//
	checkssh();

	//	
	// Ensure no related p4d processe are present.
	//
	var openthreads = threadcount(p4dname);
	if (openthreads > 0) 
	{
	    logmessage ("The " + p4dname + " processes must be stopped.\n");
	    WScript.Quit(1);
	}

	//
	// Create the directories
	//

	var dirlist = new Array(P4ROOT, P4JOURNALdir, P4LOGdir);

	logmessage("Creating directories as needed...")

	for (var n = 0; n < dirlist.length; n++)
	{
	    var mkdirret = makedir(dirlist[n]);
	}
	logmessage (" done.\n");

	//
	// Populate the db.* files by recovering from a checkpoint.
	//

	logmessage ("Replaying checkpoint...");
	var start = getseconds();

	var p4cmd = ("\"" + p4d + "\" -r \"" + P4ROOT + "\" -z -jr \"" + ZCKP + "\"");

	var oShell = new ActiveXObject("WScript.Shell");
	
	var oExec = oShell.exec(p4cmd);
	debugmessage(p4cmd);

	// Wait until the uncompress is completed until checking ExitCode
	while (oExec.Status == 0)
	{
	    WScript.Sleep(1000);
	}

	var elapsed = getelapsed(start);
	logmessage(" done.\n");
	logmessage("  checkpoint replay duration: " + parseInt(elapsed) + 
		" seconds\n");

	var ckppath = ZCKP.split('\\');
	var ckpname = ckppath[ckppath.length-1];

	writeyamlfile(yamlfile, yamllevel, "checkpointName: " + ckpname);
	writeyamlfile(yamlfile, yamllevel, "checkpointReplay: " + parseInt(elapsed));

	//
	// Copy the license file into place.
	//
	logmessage("Copying license...");
	fs = new ActiveXObject("Scripting.FileSystemObject"); 
	if (fs.FileExists(LICENSE))
	{
	    fs.CopyFile(LICENSE, P4ROOT + "\\"); 
	}

	logmessage(" done.\n");

	// Start the local server.  It will remain running throughout 
	// test and shutdown upon cleanup.
	//
	startserver();
}


function iswindows(platform)
{
	//
       	// Return true if the passed platform is Windows.
	//
	var nt = new RegExp(/^nt*/);
	return nt.test(platform);
}

function iscygwin(platform)
{
	//
	// Return true if the passed platform is Cygwin.
	//
	var cygwin = new RegExp(/^cygwin*/);
	return cygwin.test(platform);
}

function childfile(platform)
{
	//
	// Return the name of the child binary for the passed platform.
	//
	var name = CHILD;
	if (iswindows(platform) || iscygwin(platform))
	{
	    name = name + ".exe";
	}

	return name;
}

function remotechild(platform)
{
	//
	// Return the absolute path to the child binary on the passed
	// remote platform. The absolute path to the child binary is based
        // upon a temporary location commonly used on the remote platform, and
        // qualified by the name of the local host and the PID of this script
	// execution (in case the remote host is used concurrently by more
	// than one execution of this script, perhaps from more than one host).
       	// 
	//
	if (iswindows(platform) || iscygwin(platform)) 
	{
	    child = "\"\\\"" + TMPWINDOWS + CHILD + "." + getlocalhost() + "." +
			progpid + ".exe\\\"\"";
	} else {
	    child = "\"" + TMPPOSIX + CHILD + "." + getlocalhost() + "." +
			progpid + "\"";
	}
	return child;
}

function copychildfiles()
{
	logmessage("Copying child files to hosts...");

	//
	// Copy (using scp) the platform-specific child binary to each of
	// the remote hosts. The child binaries are copied from the $BINPATH
	// platform-specific directories on the local host to a temporary
	// location on the remote hosts.
	//
	var hostrecords = HOSTS.split(/ /);
	for (var r = 0; r < hostrecords.length; r++)
	{

	    //
	    //  Extract the remote hosts and platform from HOSTS collection
	    //
	    var host=getremotehost(hostrecords[r]);
	    var platform=getplatform(hostrecords[r]);

	    scpcmd = ("\"" + scp + "\" -q " + SOPTIONS + " -i" + SIDENTITY + 
		" \"" + BINPATH + platform + "\\" + childfile(platform) +  
		"\" \"" + SUSER + "@" + host + "\":" + remotechild(platform));
	    debugmessage(scpcmd);

	    //
	    // copy browsechild to remote clients
	    //
	    var oShell = new ActiveXObject("WScript.Shell");
	    var oExec = oShell.exec(scpcmd);
	
	    //
	    // Check to see whether the exit status was successful.
	    //
	    if (oExec.ProcessID == 0 && oExec.Status == 1)
	    {
		logmessage("oShell.exec failed.\n");
	    }

	    //
	    // Wait until the exec is completed before continuing.
	    //
	    while (oExec.Status == 0)
	    {
		WScript.Sleep(100);
	    }

	    // adding chmod here to workaround copssh bug
	    sshcmd = ("\"" + ssh + "\" -q " + SOPTIONS + " -i" + SIDENTITY +
		" \"" + SUSER + "@" + host + "\"" + " \"" + chmod + " " + perms + 
		"\" " + remotechild(platform));
	    debugmessage(sshcmd);

	    // chmod file to runnable permissions
	    var oExec = oShell.exec(sshcmd);

	    //
	    // Check to see whether the exit status was successful.
	    //
	    if (oExec.ProcessID == 0 && oExec.Status == 1)
	    {
		logmessage("oShell.exec failed.\n");
	    }

	    //
	    // Wait until the exec is completed before continuing.
	    //
	    while (oExec.Status == 0)
	    {
		WScript.Sleep(100);
	    }

	}

        logmessage(" done.\n");

}

function removelog()
{
	logmessage("Removing server log...");

	//
	// Remove the server log. It will be recreated by the server
	// as the children start browsing.
	//
	var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	if (oFSO.FileExists(P4LOG))
	{
	    oFSO.DeleteFile(P4LOG, true);
	}

	logmessage(" done.\n");

}

function makecommand(ichild, platform, browses, connect, seed, useroffset)
{
	//
	// Make the command to run the child.
	//
	// Start with the absolute path to the child binary on the remote 
	// platform and the port on which the server is listening.
	//
	var command = remotechild(platform) + " -p " + P4PORT;

	//
	// Append "-u user<ichild + user-offset>" (zero-filled) to the command.
	// 
	var browseuser = ichild;
	if (!isNaN(useroffset))
	{
	    browseuser = browseuser + useroffset;
	
	}

	command = command + " -u user" + getchildid(browseuser);	

	//
	// If the number of browses per child on the host was specified,
	// append "-n <browses>" to the command.
	//
        if (!isNaN(browses))
       	{ 
	    command = command + " -n " + browses;
	}

	//
	// If the type of connection each child on the host is to use while
	// browsing was specified, append "-c <connection>" to the command.
	//
        if (connect)
	{        
	    command = command + " -c " + connect;
	}


	//
	// If the base of the pseudo-random number generator seeds was specified,
	// compute the seed for this child using the passed child number on the
	// host and append "-s <computed-seed>" to the command.
	//
        if (!isNaN(seed))
	{
            command = command + " -s " +  (seed + (ichild * 173));
	}

	//
	// Return the command to run the child.
	//
        return command;

}

function waitpid(sshobj)
{
	//
	// An array containing ssh objects of all the running children are
	// passed into this function. This code iterates through
	// the objects and checks whether the stored pid is running.
	// The loop will continue until all pids are finished running.
	//
	var oWMIService = GetObject("winmgmts:\\\\.\\root\\CIMV2");

	//
	// Set the running flag to 1 to enter the loop. 
	//
	var running = 1;

	while (running != 0 )
	{
	    //
	    // Set the running flag to 0 assume process is no longer running 
	    // A running process explicitly sets flag to 1. 
	    //
	    running = 0;

	    for (var i = 0; i < sshobj.length; i++) 
	    {
		var colProcess = oWMIService.ExecQuery("SELECT processid " +
			" FROM Win32_Process WHERE ProcessId = " + sshobj[i].ProcessID);
	        for (var ei = new Enumerator(colProcess); !ei.atEnd(); ei.moveNext())
		{
		    if (ei.item().processid > 0)
		    {
			running = 1;
			debugmessage ("PROCESS " + ei.item().processid + "\n");
		    }
		} 
	    }
	//
	// Sleep for one second after iterating through ssh objects.
	//
	WScript.Sleep(1000);
	debugmessage ("----- " + running + "\n");
	}
}

function printerrors(execarray)
{
	for (var i = 0; i < execarray.length; i++)
	{
	    logmessage(execarray[i].StdOut.ReadAll());
	    logmessage(execarray[i].StdErr.ReadAll());
	}
}

function runchildren(useroffsets)
{
	var totalhosts=0;
	var totalchildren=0;

	logmessage("Running children on hosts...");

	var start=getseconds();


	//
	// Launch children on all hosts.
	//
        var hostrecords = HOSTS.split(/ /);
		
	// create an array that holds ssh objects
	var execarray = new Array();
	var index = 0;

        for (var r = 0; r < hostrecords.length; r++)
        {

	    //
	    // Gather the remotehost, platform and children count from HOSTS
	    //
	    var host=getremotehost(hostrecords[r]);
	    var platform=getplatform(hostrecords[r]);
	    var ichild=getchildren(hostrecords[r]);

	    //
	    // Gather the options that may exists in HOSTS
	    //
	    var options = getoptions(hostrecords[r]);
	    var browses = options[0];
	    var connect = options[1];
	    var seed = options[2];

	    var useroffset = useroffsets[totalhosts];

	    totalhosts++;
	    totalchildren=parseInt(totalchildren) + parseInt(ichild);


	    //
	    // For each host, launch the children from an exec'ed process.
	    // This allows each host to ramp up simultaneously.
	    //
	    while ( ichild != 0 )
	    {
		WScript.sleep(1000);

		var sshcmd = "\"" + ssh + "\" " + SOPTIONS + " -i" + SIDENTITY + 
		    " \"" + SUSER + "@" + host + "\" " +
		    makecommand(ichild, platform, browses, connect, seed, useroffset); 

		debugmessage(sshcmd);

		var oShell = new ActiveXObject("WScript.Shell");

		var oExec = oShell.exec(sshcmd);

		if (oExec.ProcessID == 0 && oExec.Status == 1)
		{
		    logmessage("oShell.exec failed.\n");
		}

		// Load an array with the oExec ssh objects. These
		// get checked against running processes in waitpid()
		execarray[index] = oExec;

		//
		// Decrement the total number of children and increase 
		// ssh objects index
		//
		ichild--;
		index++;
	    }
	}
	
	//
	// wait for the spawned ssh processes to end.
	//
	waitpid(execarray);

 	elapsed=getelapsed(start);
	
	//
	// print any errors raised from the browsechilds
	//
	printerrors(execarray);

        logmessage(" done.\n");

	//
	// Return the totalchildren, totalhosts and elapsed time.
	//
	return [totalchildren, totalhosts, elapsed];

 
}

function addelement(arr, t, c)
{
	//
	// Add an element to the array.
	//
	arr[arr.length] = [t, c];
}

function sortnumbers(a,b)
{
	//
	// This is used to sort the second elements of a two dimensional array/
	//
	return a[1] - b[1];
}

function aggregate(filename)
{
	// open the logfile
	var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	var inputlog = oFSO.OpenTextFile(localtmp + "\\" + filename, 1);

	aggarr = new Array();
	var last = "";
	var count = 0;
	var match = "";

	//
	// Aggregate the count of similar datetime strings 
	//
	while (!inputlog.AtEndOfStream)
	{
		
            line = inputlog.ReadLine();
	    var datetimestr =  line.split(" ");
	    var line  = datetimestr[0] + " " + datetimestr[1];

	    if (line != last)
	    {
		if ( count != 0 )
		{
		    addelement(aggarr, last, count);
		}
		last = line;
		count = 0;
	    } 
	    count++;
	}
	  
	if ( count != 0 )
	{ 
	    addelement(aggarr, last, count);
	}

	inputlog.Close();

	return aggarr;
}

function sortunique(filename)
{
	//
	// Open the sorted file and perform a line by line search for
	// similiar records. Write unique entries to output file.
	//
	var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	var infile = oFSO.OpenTextFile(localtmp + "\\" + filename + 
		".sort." + progpid, 1);
	var outfile = oFSO.CreateTextFile (localtmp + "\\" + filename + 
		".uniq." + progpid, true);

	var match;

	while (!infile.AtEndOfStream)
	{
	    line = infile.ReadLine();

	    if (line != match)
	    {
		match = line;
		outfile.Write(match + "\n");
	    } 
		
	}
	infile.close();
	outfile.close();

}

function logparse()
{
	var totalbrowses = 0;

	//
	// Expression to extract the date time and user string.
	//
	var re = new RegExp(/^[^0-9]*([0-9]*\/[0-9]*\/[0-9]* [0-9]*:[0-9]*:[0-9]*) pid [0-9]* (.*)@.*$/); 
	
	//
	// Expressions to find the user- entries
	//
	var userre = new RegExp(/user-/);
	var browsesre = new RegExp(/^.* 'user-filelog -L (.*)'.*$/);
	var dirsre = new RegExp(/user-dirs/);	
	var fstatre = new RegExp(/user-fstat/);	
	var filelogre = new RegExp(/user-filelog/);	
	

	//
	// Open the logfile and get ready to read it.
	//
	var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	if (oFSO.FileExists(P4LOG))
        {
	    var inputlog = oFSO.OpenTextFile(P4LOG, 1);
	} else {
	    logmessage("Log file: " + P4LOG + " doesn't exist\n");
	    logmessage("ABORTING TEST\n");
	    WScript.Quit(1);
	}	

	//
	// Create all the temp files in the local %TMP% directory.
	//
	var tmpconcurrent = oFSO.CreateTextFile(localtmp + "\\concurrent." + 
		progpid, true);
	var tmpcountbrowses = oFSO.CreateTextFile (localtmp + "\\countbrowses." +
		progpid, true);
	var tmpuserdirs = oFSO.CreateTextFile (localtmp + "\\user-dirs." + 
		progpid, true);
	var tmpuserfstat = oFSO.CreateTextFile (localtmp + "\\user-fstat." + 
		progpid, true);
	var tmpuserfilelog = oFSO.CreateTextFile (localtmp + "\\user-filelog." +
		progpid, true);

	//
	// Start reading the log. If the line matches one of the
	// regular expressions listed above, write the date/time/user
	// info to the appropriate tmp file.
	//
	while (!inputlog.AtEndOfStream)
	{
	    line = inputlog.ReadLine();
	    if (userre.test(line))
	    {
		
		var concurrentline = line.replace(re,"$1 $2");
		tmpconcurrent.Write(concurrentline + "\n");

		if (browsesre.test(line))
		{
                  var browsesline  = line.replace(browsesre,"$1");
		  tmpcountbrowses.Write(browsesline + "\n");
		  totalbrowses++
		    
		}
		if (dirsre.test(line))
		{
		  var dirsline = line.replace(re,"$1 $2");
		  tmpuserdirs.Write(dirsline + "\n");
		}
		if (fstatre.test(line))
		{
		  var fstatline = line.replace(re,"$1 $2");
		  tmpuserfstat.Write(fstatline + "\n");
		}
		if (filelogre.test(line))
		{
		  var filelogline = line.replace(re,"$1 $2");
		  tmpuserfilelog.Write(filelogline + "\n");
		}
	
	    }
	}
	//
	// Close all the open files.
	//
	tmpconcurrent.Close();
	tmpcountbrowses.Close();
	tmpuserdirs.Close();
	tmpuserfstat.Close();
	tmpuserfilelog.Close();
	inputlog.Close();

	return totalbrowses
}

function makeconcurrent()
{

        var oShell = new ActiveXObject("WScript.Shell");
        var oExec = oShell.exec("%COMSPEC% /c sort " + localtmp + 
		"\\concurrent." + progpid + " /o " + localtmp + 
		"\\concurrent.sort." + progpid + " /m " + usablemem );


        if (oExec.ProcessID == 0 && oExec.Status == 1)
	{
	    logmessage("oShell.exec failed.\n");
	}

	// Wait until command is completed before continuing.
        while (oExec.Status == 0)
	{
	    WScript.Sleep(1000);
	}

	//
	// Sort unique the results.
	//
	sortunique("concurrent");

	//
	// Return the aggregate of concurrent connections.
	//
	return aggregate("concurrent.uniq." + progpid);
}
	
function maxconcurrent(concurrent)
{
	// An array of the number of concurrent operations (timestamp and 
	// child count) are passed in where they are sorted on the 
	// second element (concurrent count) of a two dimensional array.
	// Then the array is reversed so the maximum child count is positioned
	// on the top of the array. These are assigned as the max concurrent
	// and associated timestamp.
	//
	concurrent.sort(sortnumbers);	
	concurrent.reverse();
	var tsto = concurrent[0][0];
	var max = concurrent[0][1];

	for (var i=0; i < concurrent.length; i++)
	{	
	    if (concurrent[i][1] == max)
	    {
		var tsfrom = concurrent[i][0];
	    }
		
	}

	return [max, tsfrom, tsto];
}

function getconcurrent(concurrent, timere)
{
	//
	// An array of conncurrent operations (timestamp and child count) are
	// passed in along with a timestamp to match. The array is iterated
	// through to find a matched timestamp. The number of children is 
	// returned.
	//
	for (var i=0; i < concurrent.length; i++)
	{
	    if (timere.test(concurrent[i][0]))
	    {
		children = concurrent[i][1];	
	    }
	}

	return children
}

function maxforcommand(concurrent, cmd, yamlfile)
{
	//
	// Print the maximum number of the command started in a second and the
	// date and time with the maximum number of concurrent children at
	// which the maximum number of the command started in a second
	// occurs. That is, there may be several seconds at which the same
	// maximum number of the command were started; of these, the date
	// and time printed is the first that has the maximum number of
	// concurrent children.
	//

	//
	// Sort the contents for the "command specific" logfile in the
	// tmp directory. This will result in the creation of a 
	// <command-name>.sort.<pid> file.
	//

	lfile = "user-" + cmd;

        var oShell = new ActiveXObject("WScript.Shell");
        var oExec = oShell.exec("%COMSPEC% /c sort " + localtmp + "\\" + 
		lfile + "." + progpid + " /o " + localtmp + "\\" + 
		lfile + ".sort." + progpid + " /m " +  usablemem);
        
        if (oExec.ProcessID == 0 && oExec.Status == 1)
	{
	    logmessage("oShell.exec failed.\n");
	}

	// Wait until command is completed before continuing on.
        while (oExec.Status == 0)
	{
	    WScript.Sleep(1000);
	}

	//
	// Aggregate the contents of the <command-name>.sort.<pid> file
	// based on the command count integer value.
	//

	var maxforcmd =  aggregate(lfile + ".sort." + progpid);


	//
	// Sort the results on the aggregated values and reverse the array
	// so the maximum values are on top.
	//

	maxforcmd.sort(sortnumbers);	
	maxforcmd.reverse();

	//
	// The value on the top of the array is the maximum aggregated
	// value. However, there may be other records just below it with 
	// the same max value. If they exist we need to find out whether 
	// those max'es have a higher concurrent children count.
	// We'll start by declaring the initial value at the starting 
	// max and maxts.
	//

        var element = 0;

	if ( maxforcmd.length > 0 )
	{
	    var max = maxforcmd[element][1];
	    var maxts = maxforcmd[element][0];
	    var maxtspatt = new RegExp(maxforcmd[element][0]);
	    var maxconcurrent = getconcurrent(concurrent,maxtspatt);

	    //
	    // While the subsequent records have the same "max" value
	    // we'll extract the associated max children record from the
	    //  "concurrent" array to see whether it is higher than the 
	    // maxconcurrent.
	    //
	    while (maxforcmd[element][1] == maxforcmd[element + 1][1])
	    {
		var nexttimepatt = new RegExp(maxforcmd[element + 1][0]);
		var nextconcurrent = getconcurrent(concurrent,nexttimepatt);
           
		//
		// If the nextconcurrent value is greater than the current
		// concurrent value, reassign all associated records (max,
		// time and maxconcurrent) to the new values.
		//

		if (maxconcurrent < nextconcurrent)
		{
		    max = maxforcmd[element + 1][1];
		    maxts = maxforcmd[element + 1][0];
		    maxconcurrent = nextconcurrent;
		}
		element++;

		// Don't walk off the end of the maxforcmd array. If you made
		// it this far you already have the correct timestamp for
		// the maxforcmd.
		if (element == (maxforcmd.length - 1))
		{
		    break;
		}

	    }

	    logmessage( "  maximum " + cmd + "/second: " + max);
	    logmessage( " at " + maxts +
		" (" + maxconcurrent + " concurrent children).\n");
	    writeyamlfile(yamlfile, yamllevel, "max" + chgcase(cmd) + "Second" + ": "  + max);
	    writeyamlfile(yamlfile, yamllevel, "max" + chgcase(cmd) + "Timestamp" + ": " + maxts);
	    writeyamlfile(yamlfile, yamllevel, "max" + chgcase(cmd) + "Concurrent" + ": " + maxconcurrent);
	}

}

function uniqbrowses(countbrowses)
{
        var oShell = new ActiveXObject("WScript.Shell");
        var oExec = oShell.exec("%COMSPEC% /c sort " + localtmp + "\\" + 
		countbrowses + "." + progpid + " /o " + localtmp + "\\" + 
		countbrowses + ".sort." + progpid + " /m " + usablemem);

        if (oExec.ProcessID == 0 && oExec.Status == 1)
	{
	    logmessage("oShell.exec failed.\n");
	}

	//
	// Wait until the exec is completed before continuing.
	//
        while (oExec.Status == 0)
	{
	    WScript.Sleep(1000);
	}
	
	//
	// Now we sort to unique records.
	//
	sortunique("countbrowses");

	uniquefilepaths = 0;

	// open the logfile
	var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	var inputlog = oFSO.OpenTextFile(localtmp + "\\countbrowses.uniq." + 
		progpid, 1);
	
	while (!inputlog.AtEndOfStream)
	{
	    line = inputlog.ReadLine();
	    uniquefilepaths++;
	}


	inputlog.Close();

	//
	// The length of the array indicates the number of unique paths
	//
	return uniquefilepaths;

}

function getdatestr ()
{
	//
	// Create a date object and assign variable to date components.
	//
	var oDate = new Date();
	var year = oDate.getFullYear().toString();
	var mon = (oDate.getMonth() + 1).toString();
	var day = oDate.getDate().toString();
	var hour = oDate.getHours().toString();
	var min = oDate.getMinutes().toString();
	var sec = oDate.getSeconds().toString();

	//
	// Any single position date components need to prefixed with zeros.
	//
	if(mon.length == 1) mon = "0" + mon;
	if(day.length == 1) day = "0" + day;
	if(hour.length == 1) hour = "0" + hour;
	if(min.length == 1) min = "0" + min;
	if(sec.length == 1) sec = "0" + sec;

	return (year + mon + day + hour + min + sec);
}

function getchildid(num) 
{ 
	num = num % 100000;	
	var padnum = num + '';
	while (padnum.length < 5) 
	{ 
	    padnum = '0' + padnum;
	}
	return padnum; 
}

function summarizelog(statistics)
{

        totalchildren = statistics[0];
        totalhosts = statistics[1];
        elapsed = statistics[2];

        logmessage("Summarizing server log...\n");

        logmessage("  " + totalchildren + " children over " + totalhosts +
                " hosts browsed for " + parseInt(elapsed) + " seconds.\n");

	writeyamlfile(yamlfile, yamllevel, "totalChildren: " +  totalchildren);
	writeyamlfile(yamlfile, yamllevel, "totalHosts: " +  totalhosts);
	writeyamlfile(yamlfile, yamllevel, "totalSeconds: " +  parseInt(elapsed));

	totalbrowses = logparse();

	var concurrent = makeconcurrent();

	var maxstats = maxconcurrent(concurrent);

	logmessage("  " + maxstats[0] + " children ran concurrently from " + 
		maxstats[1] + " to " + maxstats[2] + "\n");
	writeyamlfile(yamlfile, yamllevel, "concurrentChildren: " + maxstats[0]);
	writeyamlfile(yamlfile, yamllevel, "concurrentFrom: " + maxstats[1]);
	writeyamlfile(yamlfile, yamllevel, "concurrentTo: " + maxstats[2]);

        maxforcommand(concurrent, "dirs", yamlfile);

        maxforcommand(concurrent, "fstat", yamlfile);

        maxforcommand(concurrent, "filelog", yamlfile);


	logmessage("  " + totalbrowses + " total browses browsed " + 
		uniqbrowses("countbrowses") + " unique filepaths.\n");
	writeyamlfile(yamlfile, yamllevel, "totalBrowses: " + totalbrowses);
	writeyamlfile(yamlfile, yamllevel, "uniquePaths: " + uniqbrowses("countbrowses"));

	var tmpfiles = new Array("concurrent", 
				"countbrowses", 
				"user-dirs",
				"user-fstat",
				"user-filelog");

	oFSO = new ActiveXObject("Scripting.FileSystemObject");
	for (i=0; i < tmpfiles.length; i++)
	{
	    oFSO.DeleteFile(localtmp + "\\" + tmpfiles[i] + "*" + progpid);
	}

}

function removechildfiles()
{
	logmessage("Removing child files from hosts...");

	var hostrecords = HOSTS.split(/ /);
	for (var r = 0; r < hostrecords.length; r++)
	{
	    var host=getremotehost(hostrecords[r]);
	    var platform=getplatform(hostrecords[r]);
			
	    var sshcmd =  "\"" + ssh + "\" " + SOPTIONS + " -i" + SIDENTITY + " " +
	        SUSER + "@" + host + " " + rm  + " " + remotechild(platform);

	    //
	    // Remove the browsechild from remote clients.
	    //
	    var oShell = new ActiveXObject("WScript.Shell");
	    var oExec = oShell.exec(sshcmd);

	    if (oExec.ProcessID == 0 && oExec.Status == 1)
	    {
		logmessage("oShell.exec failed.\n");
	    }

	    //
	    // Wait until the exec is completed before continuing.
	    //
	    while (oExec.Status == 0)
	    {
		WScript.Sleep(100);
	    }

	}

        logmessage(" done.\n");

}

function movelog()
{

	//
	// Move the existing log file to current working directory
	// using the assembled date/time string as the naming
	// convention.
	//
	var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	var target = oFSO.GetFolder(".") + "\\" + BENCHMARK + "." + DATETIME + ".log";

        logmessage("Moving server log to " + target + "...");
	if (oFSO.FileExists(P4LOG))
	{
	    oFSO.MoveFile(P4LOG, target);
	}

	logmessage(" done.\n");
}

function runme()
{
	//
	// Check ssh connectivity between the server and client machines
	// again; bad things might have happened since the setup phase
	// if it was executed separately.
	//
	checkssh();

	//
	// Ensure that the server is listening.
	//
        if (islistening() != 0)
	{
            logmessage("Perforce server is not listening!\n");
            logmessage("Run the \"setup\" phase if it has not been run," +
                " or run the \"startserver\" phase.");
	    WScript.Quit(1);
	}	

	//
	// Get the number of children specified (or the default) and
	// the user offsets specified in the HOSTS definition.
	//
	var nhosts = 0
	var nchildrens = new Array();
	var useroffsets = new Array();
	var hostrecords = HOSTS.split(/ /);
	for (var r = 0; r < hostrecords.length; r++)
	{
	    var raw = hostrecords[r];
	    nchildrens[nhosts] = getchildren(raw);
	    useroffsets[nhosts] = getuseroffset(raw);
	    nhosts++;
	}

	//
	// Ensure that the number of children and user offsets specified
	// in the HOSTS definition will result in unique users.
	//
	if (!usersfit(nhosts, nchildrens, useroffsets))
	{
	    logmessage("Users in the HOSTS definition overlap!\n");
	    logmessage("Correct the number of children and/or the\n");
	    logmessage("user offsets in the HOSTS definition\n");
	    logmessage("and restart the \"runme\" phase.\n");
	    WScript.Quit(1);
	}

	//
	// Make user offsets for those that were not specified in
	// the HOSTS definition such that all resulting users will be unique.
	//
	fituseroffsets(nhosts, nchildrens, useroffsets);

        //
        // Copy child files to temporary locations on the hosts.
        //
	copychildfiles();

        //
        // Remove the server log so that the summarized results
        // reflect only this run.
        //
	removelog();

        //
        // Run children on the hosts.
        //
        var statistics = runchildren(useroffsets);

        //
        // Summarize the results for this run.
        //
        summarizelog(statistics);

        //
        // Remove child files from the hosts.
        //
        removechildfiles();

        //
	// Save the server log for later review.
        //
        movelog();

}

function removefolder(oFSO, folder)
{
	oFSO.DeleteFolder(folder);
}

function cleanup()
{
	//
	// Stop the server and remove the db*, license
	stopserver();

	logmessage("Removing db.* files, license, journal file, and server log...");
	var oFSO = new ActiveXObject("Scripting.FileSystemObject");
	fo = oFSO.GetFolder(P4ROOT);
	ff = new Enumerator(fo.Files);


	//
	// Regular expression to find all db.* files under P4ROOT.
	//
	var dbre = new RegExp(/db\..*/);

	// Iterate through the collection of files found in P4ROOT
	// and delete them if they match db.*
	
	for (; !ff.atEnd(); ff.moveNext())	
	{
	    var ofile = ff.item();
	    if (dbre.test(ofile))
	    {
		debugmessage("DELETE " + ofile);
		oFSO.DeleteFile(ofile);
	    }
	}        
		 
	rmfiles = new Array(    P4ROOT + "\\license",
				P4JOURNAL,
				P4LOG);

	for (i=0; i < rmfiles.length; i++)
	{
	    if (oFSO.FileExists(rmfiles[i]))
	    {
		debugmessage("DELETE " + rmfiles[i]);
		oFSO.DeleteFile(rmfiles[i]);
	    }
	}
        logmessage(" done.\n");

	if (oFSO.FolderExists(P4ROOT + "\\server.locks"))
	{
	    logmessage("Removing server.locks subdirectory...")
	    removefolder(oFSO, P4ROOT + "\\server.locks")
	    logmessage(" done.\n");
	}

}

// If a p4s.exe process is running raise error.
var tcnt = threadcount("p4s.exe");
if (tcnt)
{
    logmessage("p4s.exe should be stopped before running test\n");
    WScript.Quit(1);
}

// If no arguments are given use "setup runme cleanup"
var argcnt = WScript.Arguments.length;
var phase = new Array();

if (argcnt == 0)
{
    phase = Array("setup", "runme", "cleanup");
} else {
    phase[0] = WScript.Arguments(0);

}

var rcschange =  getrcsvalue(RCSCHANGE);
var rcsdate =  getrcsvalue(RCSDATE);
p4dver = getversion(p4d);

// Display the browse.js and p4* version information.
logmessage("Using versions:\n");
logmessage("  BROWSE.JS/" + rcschange +
	" (" + rcsdate + ")\n");
logmessage("  " + p4dver + "\n");
logmessage("On machine:\n");
logmessage("  " + uname + " " + SUSER + " "  + os + " "  + arch + "\n");

var yamlfile = createyamlfile();
logmessage("BRDB file:\n");
logmessage("  " +  yamlfile + "\n");

var yamllevel = 0;
writeyamlfile(yamlfile, yamllevel, "benchmark: " + BENCHMARK);
writeyamlfile(yamlfile, yamllevel, "args:" + getargs());
writeyamlfile(yamlfile, yamllevel, "testInfo:");
yamllevel = yamllevel + 1;
writeyamlfile(yamlfile, yamllevel, "rcsChange: " + getrcsvalue(RCSCHANGE));
writeyamlfile(yamlfile, yamllevel, "rcsDate: " + getrcsvalue(RCSDATE));
yamllevel = yamllevel - 1;
writeyamlfile(yamlfile, yamllevel, "versions:");
yamllevel = yamllevel + 1;
writeyamlfile(yamlfile, yamllevel, "p4d: " + p4dver);
yamllevel = yamllevel - 1;
writeyamlfile(yamlfile, yamllevel, "uname: "   + "\"" + uname + " " + SUSER + " "  + os + " "  + arch + "\"");

var oFSO = new ActiveXObject("Scripting.FileSystemObject");
var t = oFSO.GetFolder(".") + "\\" + WScript.ScriptName;
var fname = t.replace(/\\/g,"\\\\");


// Set the environment variables.
setenv();

var progpid = getpid();
var usablemem = getusablemem();

var irunme = 0;

for (var n = 0; n < phase.length; n++)
{
    logmessage("\nStarting " + BENCHMARK + " " + phase[n] + " phase...\n");
    switch (phase[n])
    {
	case "setup":
	    setup(yamlfile, yamllevel);
	    break;
	case "runme":
	    if (irunme == 0)
	    {
		writeyamlfile(yamlfile, yamllevel, "runme:");
		yamllevel = yamllevel + 1;
	    }
	    irunme++;
	    writeyamlfile(yamlfile, yamllevel, irunme + ":");
	    yamllevel = yamllevel + 1;
	    runme();
	    yamllevel = yamllevel - 1;
	    break;
	case "cleanup":
	    cleanup();
	    break;
	case "checkssh":
	    checkssh();
	    break;
	case "startserver":
	    startserver();
	    break;
	case "stopserver":
	    stopserver();
	    break;
	default:
	    WScript.echo("\nPhase " + phase[n] + " unknown! Terminating.\n");
	    WScript.Quit(1);
    }
}
